/****************************************************************
 * Copyright 2004 All rights reserved                                *
 * HI Corporation and Sony Ericsson Mobile Communications AB         *
 *                                                                   *
 * The software is the copyrighted work of HI Corporation with       *
 * modifications from Sony Ericsson Mobile Communications AB.        *
 *                                                                   *
 * The use of the software is subject to the terms of the            *
 * end-user license agreement which accompanies or is included       *
 * with the software. The software is provided "as is" and Sony      *
 * Ericsson specifically disclaim any warranty or condition          *
 * whatsoever regarding merchantability or fitness for a             *
 * specific purpose, title or non-infringement. No warranty of       *
 * any kind is made in relation to the condition, suitability,       *
 * availability, accuracy, reliability, merchantability and/or       *
 * non-infringement of the software provided herein                  *
 ****************************************************************/

import com.mascotcapsule.micro3d.v3.ActionTable;
import com.mascotcapsule.micro3d.v3.AffineTrans;
import com.mascotcapsule.micro3d.v3.Effect3D;
import com.mascotcapsule.micro3d.v3.Figure;
import com.mascotcapsule.micro3d.v3.FigureLayout;
import com.mascotcapsule.micro3d.v3.Graphics3D;
import com.mascotcapsule.micro3d.v3.Light;
import com.mascotcapsule.micro3d.v3.Texture;
import com.mascotcapsule.micro3d.v3.Vector3D;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.game.GameCanvas;

/**
 * Loads, displays and rotates 3d models on a Canvas.
 * The Model has two different animations. Switch between
 * these by soft key "Change" and rotate the
 * world/models with the joystick.
 */
public class Canvas3D extends GameCanvas implements CommandListener
{

    /**
     * Rotation value expressed in a format
     * where 360 degrees equals 4096 units.
     */
    private static final int rotationIncrement = 5;

    /**
     * Midlet
     *
     */
    private MascotCapsuleDemo app;

    /**
     * 3D model
     */
    private static Texture     backgroundTexture;
    private static Texture     robotTexture;
    private static Figure      backgroundFigure;
    private static Figure      robotFigure;
    private static ActionTable robotAction;

    /**
     * The current frame in the animation
     */
    private static int currentFrame = 0;

    /**
     * There are two different animations that is loaded
     * dynamically using the same variable (robotAction).
     * robotActionNumber is used to know which one we are currently dealing with.
     */
    private static int robotActionNumber = 1;

    /**
     * 3D
     */
    private Vector3D BACKGROUND_POSITION    = new Vector3D(0,0,450);
    private Vector3D BACKGROUND_LOOK        = new Vector3D(0,-500,-4096);
    private Vector3D BACKGROUND_UP          = new Vector3D(0,4096,0);
    private int      BACKGROUND_PERSPECTIVE = 512;

    private Vector3D ROBOT_POSITION         = new Vector3D(0,0,450);
    private Vector3D ROBOT_LOOK             = new Vector3D(0,-500,-4096);
    private Vector3D ROBOT_UP               = new Vector3D(0,4096,0);
    private int      ROBOT_PERSPECTIVE      = 512;

    /**
     * Light
     *
     */
    private Vector3D lightDirVec = new Vector3D(-146,293,439);
    private int lightDir = 3730;
    private int lightAmbi = 1626;
    private Light    backgroundLight = new Light(lightDirVec,lightDir,lightAmbi);

    // AffineTrans for background
    private AffineTrans backgroundXTrans;
    private AffineTrans backgroundYTrans;
    private AffineTrans backgroundZTrans;
    private AffineTrans backgroundTrans;
    private AffineTrans backgroundVTrans;

    // AffineTrans for robot
    private AffineTrans robotXTrans;
    private AffineTrans robotYTrans;
    private AffineTrans robotZTrans;
    private AffineTrans robotTrans;
    private AffineTrans robotVTrans;

    /**
     * Initial rotation position (3D model facing the user).
     * Note that we rotate the model instead of the camera!
     */
    private int rotationX = 100;
    private int rotationY = 0;

    private FigureLayout backgroundLayout;
    private FigureLayout robotLayout;
    private Effect3D     effect;

    /**
     * Soft keys
     */
    private Command changeCommand;
    private Command exitCommand;

    /**
     * Graphics3D object cached by getGraphics3D() method
     */
    private Graphics3D cachedGraphics3D;

    /**
     * Create new Canvas3D for MascotCapsuleDemo
     */
    Canvas3D(MascotCapsuleDemo app)
    {
        super(false);
        this.app = app;

        // Load the Texture
        try
        {
          backgroundTexture = loadTexture("/tex_001.bmp",true);
        }
        catch (Exception e)
        {
          e.printStackTrace();
        }

        // Loads the Figures
           // backgroundFigure is the 3D background
           // robotFigure is the robot
        try
        {
          backgroundFigure = loadFigure("/test_model_haikei.mbac");
          robotFigure      = loadFigure("/test_model_robo.mbac");
        }
        catch (Exception e)
        {
          e.printStackTrace();
        }

        // Sets the Texture
        backgroundFigure.setTexture(backgroundTexture);
        robotFigure.setTexture(backgroundTexture);

        // Loads the first action
        try
        {
          robotAction = loadActionTable("/action_01.mtra");
        }
        catch (Exception e)
        {
          e.printStackTrace();
        }


        // 3D Init
        effect = new Effect3D();

        // Set the shading and lights
        effect.setShadingType(Effect3D.NORMAL_SHADING);
        effect.setLight(backgroundLight);

        // Init the Soft keys
        changeCommand = new Command("Change",Command.SCREEN,1);
        exitCommand = new Command("Quit",Command.EXIT,1);
        addCommand(exitCommand);
        addCommand(changeCommand);
        setCommandListener(this);

        // Background Affine Transformation variables --
        backgroundXTrans = new AffineTrans();
        backgroundYTrans = new AffineTrans();
        backgroundZTrans = new AffineTrans();
        backgroundTrans  = new AffineTrans();
        backgroundVTrans = new AffineTrans();

        // Robot Affine Transformation variables --
        robotXTrans = new AffineTrans();
        robotYTrans = new AffineTrans();
        robotZTrans = new AffineTrans();
        robotTrans  = new AffineTrans();
        robotVTrans = new AffineTrans();

        // Background layout
        backgroundLayout = new FigureLayout();
        backgroundLayout.setPerspective(1,4096,BACKGROUND_PERSPECTIVE);
        backgroundLayout.setCenter(getWidth()/2, getHeight()/2);

        // Robot layout
        robotLayout = new FigureLayout();
        robotLayout.setPerspective(1,4096,ROBOT_PERSPECTIVE);
        robotLayout.setCenter(getWidth()/2, getHeight()/2);

    }

    /**
     * Create Graphics3D object by demand
     * and use cached value for subsequent calls
     */
    private Graphics3D getGraphics3D()
    {
      if (cachedGraphics3D == null)
      {
        cachedGraphics3D = new Graphics3D();
      }
      return cachedGraphics3D;
    }

    /**
     * Calculates the new position for the 3D models and paints them
     * on the 2D display
     */
    synchronized void processPaint()
    {
        Graphics g = getGraphics();
        Graphics3D g3d = getGraphics3D();

        g.setColor(0,0,0);
        g.fillRect(0,0,getWidth(),getHeight());

        /**
         * Background transformation
         */
        backgroundXTrans.setIdentity();
        backgroundYTrans.setIdentity();
        backgroundZTrans.setIdentity();
        backgroundTrans.setIdentity();

        backgroundXTrans.rotationX(rotationX);  // X-axis rotation
        backgroundYTrans.rotationY(rotationY);  // Y-axis rotation
        backgroundZTrans.rotationZ(0);          // We don't do any z-axis rotation
        backgroundYTrans.mul(backgroundXTrans);
        backgroundZTrans.mul(backgroundYTrans);
        backgroundVTrans.setIdentity();

        /**
         * Background lookAt
         */
        backgroundVTrans.lookAt(BACKGROUND_POSITION, BACKGROUND_LOOK, BACKGROUND_UP);
        backgroundTrans.mul(backgroundVTrans, backgroundZTrans);
        backgroundLayout.setAffineTrans(backgroundTrans);

        /**
         * Robot transformation
         */
        robotXTrans.setIdentity();
        robotYTrans.setIdentity();
        robotZTrans.setIdentity();
        robotTrans.setIdentity();

        robotXTrans.rotationX(rotationX);  // X-axis rotation
        robotYTrans.rotationY(rotationY);  // Y-axis rotation
        robotZTrans.rotationZ(0);          // We don't do any z-axis rotation
        robotYTrans.mul(robotXTrans);
        robotZTrans.mul(robotYTrans);
        robotVTrans.setIdentity();

        /**
         * Robot lookAt
         */
        robotVTrans.lookAt(ROBOT_POSITION, ROBOT_LOOK, ROBOT_UP);
        robotTrans.mul(robotVTrans, robotZTrans);
        robotLayout.setAffineTrans(robotTrans);

        /*
         * Calculate what frame to show and set the posture accordingly.
         * Note that getNumFrames(0) returns a number that is very much
         * higher than the number of frames created in the 3D modelling tool
         * This means we can increase or decrease the animation speed by chosing
         * how many frames of the total number of frames from getNumFrames(0) we
         * want to show, before the animations starts from the begining again.
         */
        currentFrame += robotAction.getNumFrames(0)/30;

        if(currentFrame > robotAction.getNumFrames(0))
        {
            currentFrame = 0;
        }
        robotFigure.setPosture(robotAction,0,currentFrame);

        try
        {
            g3d.bind(g);

            g3d.renderFigure(backgroundFigure,0,0,backgroundLayout,effect);
            g3d.renderFigure(robotFigure,0,0,robotLayout,effect);
            g3d.flush();

            g3d.release( g );

        }
        catch(Throwable h)
        {
            h.printStackTrace();
        }
        flushGraphics();
    }

    /**
     * Loads the Figure from the mbac file
     */
    private Figure loadFigure(String fileName) throws Exception
    {
        return new Figure(fileName);
    }

    /**
     * Loads the ActionTable from the mtra file
     */
    private ActionTable loadActionTable(String fileName) throws Exception
    {
        return new ActionTable(fileName);
    }

    /**
     * Loads the Texture from the bmp file
     */
    private Texture loadTexture(String fileName, boolean noSphere) throws Exception
    {
        return new Texture(fileName, noSphere);
    }

    /**
     * Checks if midlet is still running
     */
    boolean resourcesAreLoaded()
    {
      return ((backgroundFigure != null)
              && (robotFigure != null)
              && (backgroundTexture != null)
              && (robotAction != null));
    }

    /**
     * Clear all resources
     */
    void clearResource()
    {
        if( backgroundFigure != null )
        {
            backgroundFigure.dispose();
            backgroundFigure = null;
        }
        if( robotFigure != null )
        {
            robotFigure.dispose();
            robotFigure = null;
        }
        if( backgroundTexture != null )
        {
            backgroundTexture.dispose();
            backgroundTexture = null;
        }
        if( robotAction != null )
        {
            robotAction.dispose();
            robotAction = null;
        }
    }

    /**
     * Sets the rotation of the models/world according to what
     * keys/joystick are moved
     */
    synchronized void processKeyState()
    {
        int keyStates = getKeyStates();
        // UP
        if( (keyStates & UP_PRESSED) != 0)
        {
            rotationX += rotationIncrement;
            if( rotationX > 4096 )
            {
                rotationX = 0;
            }
        }
        // DOWN
        if( (keyStates & DOWN_PRESSED) != 0) {
            rotationX -= rotationIncrement;
            if(rotationX < 0 )
            {
                rotationX = 4096;
            }
        }
        // LEFT
        if( (keyStates & LEFT_PRESSED) != 0)
        {
            rotationY += rotationIncrement;
            if( rotationY > 4096 )
            {
                rotationY = 0;
            }
        }
        // RIGHT
        if( (keyStates & RIGHT_PRESSED) != 0)
        {
            rotationY -= rotationIncrement;
            if(rotationY < 0 )
            {
                rotationY = 4096;
            }
        }
        // FIRE - sets the figure to its initial position
        if( (keyStates & FIRE_PRESSED) != 0)
        {
            rotationX = 100;
            rotationY = 0;
        }
    }

    /**
     * Handles the soft keys.
     * Left soft key changes the animation,
     * right soft key quits the MIDlet.
     */
    synchronized public void commandAction(Command command, Displayable displayable)
    {
        if(command == changeCommand)
        {
            if( robotActionNumber == 1 )
            {
                robotActionNumber = 2;
            }
            else
            {
                robotActionNumber = 1;
            }
            currentFrame = 0;
            // Disposes of the old, and loads the new action. Note that this should really be
            // done in a new Thread so execution can leave this method as soon as possible!
            try
            {
                robotAction.dispose();
                robotAction = loadActionTable("/action_0"+robotActionNumber+".mtra");
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }
        else if(command == exitCommand)
        {
            clearResource();
            app.destroyApp(true);
        }
    }
}
